
from grammar import ExtendsType, SemanticError
from . import types

class Ext (metaclass=ExtendsType):
    from .parser import CXXParser as SRCMODULE

class ExtCPP (metaclass=ExtendsType):
    from .parser import CPPParser as SRCMODULE

class Context:
    def __init__(self):
        self.globalns = types.Namespace()
        self.includes = {}

class TranslationUnit (Ext):
    def pp_main(self, ns, ctxt):
        for d in self.decls:
            d.pp_declare(ns, ctxt)

#class NameTypedef (Ext):
#    def pp_lookup(self, ns, ctxt, tmplt=False):
#        if not tmplt:
#            o = ns.lookup(self.name)
#            if isinstance(o, types.NamespaceObj): return o
#        raise SemanticError("Type name expected", self.src)

class NameType (Ext):
    def pp_lookup(self, ns, ctxt, tmplt=False):
        if not tmplt:
            o = ns.lookup(self.name)
            if ns.within(self.name):
                return o
            if isinstance(o, (types.Class, types.Typedef, types.Enum,
                              types.TypeParam)): return o
        raise SemanticError("Type name expected", self.src)

class NameClass (Ext):
    def pp_lookup(self, ns, ctxt, tmplt=False):
        if not tmplt:
            o = ns.lookup(self.name)
            if ns.within(self.name):
                return o
            if isinstance(o, types.Class): return o
        raise SemanticError("Class name expected", self.src)

    def pp_nslookup(self, ns, ctxt, tmplt=False):
        return self.pp_lookup(ns, ctxt, tmplt=tmplt).members

#class NameEnum (Ext):
#    def pp_lookup(self, ns, ctxt, tmplt=False):
#        if not tmplt:
#            o = ns.lookup(self.name)
#            if isinstance(o, types.Enum): return o
#        raise SemanticError("Enum name expected", self.src)

class NameNamespace (Ext):
    def pp_nslookup(self, ns, ctxt, tmplt=False):
        if not tmplt:
            o = ns.lookup(self.name)
            if isinstance(o, types.NamespaceObj): return o.nns
        raise SemanticError("Namespace name expected", self.src)

#class NameNamespaceAlias (Ext):
#    def pp_nslookup(self, ns, ctxt, tmplt=False):
#        if not tmplt:
#            o = ns.lookup(self.name)
#            if isinstance(o, types.NamespaceAlias): return o.nns
#        raise SemanticError("Namespace alias expected", self.src)

class LiteralStr (Ext):
    def pp_expr(self, ns, ctxt):
        pass

class LiteralChar (Ext):
    def pp_expr(self, ns, ctxt):
        pass

class LiteralInt (Ext):
    def pp_expr(self, ns, ctxt):
        pass

class LiteralFloat (Ext):
    def pp_expr(self, ns, ctxt):
        pass

class LiteralBool (Ext):
    def pp_expr(self, ns, ctxt):
        pass

class NestedNameSpecifier (Ext):
    def pp_nslookup(self, ns, ctxt, tmplt=False):
        o = self.name.pp_nslookup(ns, ctxt, tmplt=tmplt)
        if self.nns is None: return o
        return self.nns.pp_nslookup(o, ctxt, tmplt=self.t)

class IdTemplate (Ext):
    def pp_nslookup(self, ns, ctxt, tmplt=False):
        o = ns.lookup(self.name)
        if not isinstance(o, types.TemplateClass):
            raise SemanticError("Template class name expected", self.src)
        for a in self.args:
            a.pp_tmpltarg(ns, ctxt)
        return o.members

    def pp_lookup(self, ns, ctxt, tmplt=False):
        o = ns.lookup(self.name)
        if not isinstance(o, types.TemplateClass):
            raise SemanticError("Template class name expected", self.src)
        for a in self.args:
            a.pp_tmpltarg(ns, ctxt)
        return o

class TmpltArgValue (Ext):
    def pp_tmpltarg(self, ns, ctxt):
        self.v.pp_expr(ns, ctxt)

class TmpltArgType (Ext):
    def pp_tmpltarg(self, ns, ctxt):
        self.t.pp_type(ns, ctxt)

class IdQualified (Ext):
    def pp_expr(self, ns, ctxt):
        if self.glbl: ns = ctxt.globalns
        if self.nns is not None: ns = self.nns.pp_nslookup(ns, ctxt)
        return self.id.pp_lookup(ns, ctxt, tmplt=self.t)

    def pp_attr(self, ns, ctxt):
        if self.glbl: ns = ctxt.globalns
        if self.nns is not None: ns = self.nns.pp_nslookup(ns, ctxt)

    def pp_dtor(self, decl, ns, ctxt):
        if self.glbl: ns = ctxt.globalns
        if self.nns is not None: ns = self.nns.pp_nslookup(ns, ctxt)
        o = self.id.pp_lookup(ns, ctxt, tmplt=self.t)
        if o is None or isinstance(o, types.FunctionDef):
            raise SemanticError("Bad declarator", self.src)
        return o

class IdIdentifier (Ext):
    def pp_lookup(self, ns, ctxt, tmplt=False):
        if not tmplt:
            o = ns.lookup(self.id)
            if isinstance(o, (types.Variable, types.FunctionDef)): return o
        raise SemanticError("Variable/Function name expected", self.src)

    def pp_attr(self, ns, ctxt):
        pass

    def pp_dtor(self, decl, ns, ctxt):
        return decl.declare(self.id, ns, ctxt, self.src)

    def pp_expr(self, ns, ctxt):
        o = ns.lookup(self.id)
        if o is None:
            raise SemanticError("Undefined Variable", self.src)
        if not isinstance(o, (types.Variable, types.FunctionDef)):
            raise SemanticError("Object is not a Variable", self.src)

class IdOperatorFn (Ext):
    def pp_dtor(self, decl, ns, ctxt):
        return decl.declare('+op'+self.op, ns, ctxt, self.src)

    def pp_attr(self, ns, ctxt):
        pass

class IdConversionFn (Ext):
    def pp_dtor(self, decl, ns, ctxt):
        decl = types.Declare()
        self.type.pp_dtype(decl, ns, ctxt)
        if self.dtor:
            decl = self.dtor.pp_dtor(decl, ns, ctxt)
        return decl.declare('+conv+', ns, ctxt, self.src)

class ExprThis (Ext):
    def pp_expr(self, ns, ctxt):
        pass

class ExprDelete (Ext):
    def pp_expr(self, ns, ctxt):
        self.v.pp_expr(ns, ctxt)

class ExprNew (Ext):
    def pp_expr(self, ns, ctxt):
        if self.place:
            for p in self.place:
                p.pp_expr(ns, ctxt)
        self.t.pp_type(ns, ctxt)
        if self.init:
            for p in self.init:
                p.pp_expr(ns, ctxt)

class NewTypeId (Ext):
    def pp_type(self, ns, ctxt):
        decl = types.Declare()
        self.type.pp_dtype(decl, ns, ctxt)
        if self.dtor:
            return self.dtor.pp_dtor(decl, ns, ctxt)

class ExprCondition (Ext):
    def pp_expr(self, ns, ctxt):
        self.a.pp_expr(ns, ctxt)
        self.b.pp_expr(ns, ctxt)
        self.c.pp_expr(ns, ctxt)

class BinopNode (Ext):
    def pp_expr(self, ns, ctxt):
        self.a.pp_expr(ns, ctxt)
        self.b.pp_expr(ns, ctxt)

class ExprAssign (Ext):
    def pp_expr(self, ns, ctxt):
        self.lhs.pp_expr(ns, ctxt)
        self.rhs.pp_expr(ns, ctxt)

class UnopNode (Ext):
    def pp_expr(self, ns, ctxt):
        self.v.pp_expr(ns, ctxt)

class ExprAttributeAccess (Ext):
    def pp_expr(self, ns, ctxt):
        self.v.pp_expr(ns, ctxt)
        self.id.pp_attr(ns, ctxt)

class ExprIndirectAttributeAccess (Ext):
    def pp_expr(self, ns, ctxt):
        self.v.pp_expr(ns, ctxt)
        self.id.pp_attr(ns, ctxt)

class ExprSizeof (Ext):
    def pp_expr(self, ns, ctxt):
        self.v.pp_expr(ns, ctxt)

class ExprSizeofType (Ext):
    def pp_expr(self, ns, ctxt):
        self.t.pp_type(ns, ctxt)

class ExprThrow (Ext):
    def pp_expr(self, ns, ctxt):
        self.v.pp_expr(ns, ctxt)

class ExprCCast (Ext):
    def pp_expr(self, ns, ctxt):
        self.t.pp_type(ns, ctxt)
        self.v.pp_expr(ns, ctxt)

class ExprPtrToMembDereference (Ext):
    def pp_expr(self, ns, ctxt):
        self.a.pp_expr(ns, ctxt)
        self.b.pp_expr(ns, ctxt)

class ExprIndirectPtrToMembDereference (Ext):
    def pp_expr(self, ns, ctxt):
        self.a.pp_expr(ns, ctxt)
        self.b.pp_expr(ns, ctxt)

class ExprIndex (Ext):
    def pp_expr(self, ns, ctxt):
        self.v.pp_expr(ns, ctxt)
        self.i.pp_expr(ns, ctxt)

class ExprCall (Ext):
    def pp_expr(self, ns, ctxt):
        self.v.pp_expr(ns, ctxt)
        for a in self.args:
            a.pp_expr(ns, ctxt)

class ExprCreate (Ext):
    def pp_expr(self, ns, ctxt):
        self.t.pp_type(ns, ctxt)
        for a in self.args:
            a.pp_expr(ns, ctxt)

class StmtCompound (Ext):
    def pp_stmt(self, ns, ctxt):
        lns = types.Namespace(ns)
        for s in self.sts:
            s.pp_stmt(lns, ctxt)

class StmtExpression(Ext):
    def pp_stmt(self, ns, ctxt):
        self.expr.pp_expr(ns, ctxt)

class StmtReturn (Ext):
    def pp_stmt(self, ns, ctxt):
        if self.v:
            self.v.pp_expr(ns, ctxt)

class StmtIf (Ext):
    def pp_stmt(self, ns, ctxt):
        self.cond.pp_expr(ns, ctxt)
        self.thenst.pp_stmt(ns, ctxt)
        if self.elsest:
            self.elsest.pp_stmt(ns, ctxt)

class StmtSwitch (Ext):
    def pp_stmt(self, ns, ctxt):
        self.cond.pp_expr(ns, ctxt)
        self.st.pp_stmt(ns, ctxt)

class StmtCase (Ext):
    def pp_stmt(self, ns, ctxt):
        self.c.pp_expr(ns, ctxt)
        self.st.pp_stmt(ns, ctxt)

class StmtDefault (Ext):
    def pp_stmt(self, ns, ctxt):
        self.st.pp_stmt(ns, ctxt)

class StmtWhile (Ext):
    def pp_stmt(self, ns, ctxt):
        self.cond.pp_expr(ns, ctxt)
        self.st.pp_stmt(ns, ctxt)

class StmtFor (Ext):
    def pp_stmt(self, ns, ctxt):
        lns = types.Namespace(ns)
        self.init.pp_stmt(lns, ctxt)
        if self.cond:
            self.cond.pp_expr(lns, ctxt)
        if self.inc:
            self.inc.pp_expr(lns, ctxt)
        self.st.pp_stmt(lns, ctxt)

class StmtTry (Ext):
    def pp_stmt(self, ns, ctxt):
        self.body.pp_stmt(ns, ctxt)
        for h in self.handlers:
            h.pp_stmt(ns, ctxt)

class StmtCatch (Ext):
    def pp_stmt(self, ns, ctxt):
        lns = types.Namespace(ns)
        self.exc.pp_except(lns, ctxt)
        self.body.pp_stmt(lns, ctxt)

class DeclException (Ext):
    def pp_except(self, ns, ctxt):
        decl = types.Declare()
        self.type.pp_dtype(decl, ns, ctxt)
        decl.astnode = ([], self.type, self.dtor)
        self.dtor.pp_dtor(decl, ns, ctxt)

class StmtDeclaration (Ext):
    def pp_stmt(self, ns, ctxt):
        self.decl.pp_declare(ns, ctxt)

class TypeTypename (Ext):
    def pp_dtype(self, decl, ns, ctxt):
        if self.glbl: ns = ctxt.globalns
        if self.nns is not None: ns = self.nns.pp_nslookup(ns, ctxt)
        o = ns.lookup(self.name)
        if isinstance(o, types.TypeObject):
            if decl: decl.set_type(o)
        elif decl:
            decl.set_type(types.TypeParam("<anon>"))

class TypeClass (Ext):
    def pp_declare(self, ns, ctxt):
        self.pp_dtype(None, ns, ctxt)

    def pp_dtype(self, decl, ns, ctxt):
        if self.glbl: ns = ctxt.globalns
        if self.nns is not None: ns = self.nns.pp_nslookup(ns, ctxt)
        o = ns.lookup(self.name)
        if isinstance(ns, types.Memberspace) and self.name == ns.clsname:
            if decl: decl.set_type(o)
        elif isinstance(o, types.Class):
            if decl: decl.set_type(o)
        elif self.glbl or self.nns:
            raise SemanticError("Class name expected", self.src)
        else:
            ns.dict[self.name] = types.Class(self.name, None, None)

    def pp_template(self, plist, ns, ctxt):
        if self.glbl or self.nns:
            raise SemanticError("May only declare in local namespace", self.src)
        o = ns.lookup(self.name)
        if o is None:
            ns.dict[self.name] = types.TemplateClass(self.name, None, None)
        elif not isinstance(o, types.TemplateClass):
            raise SemanticError("Incompatible declaration of "+self.name,
                                self.src)

class DeclLinkage (Ext):
    def pp_declare(self, ns, ctxt):
        for s in self.body:
            s.pp_declare(ns, ctxt)

class DeclFunction (Ext):
    def pp_declare(self, ns, ctxt):
        return self.head.pp_declare_fn(self.init, self.body, ns, ctxt)

    def pp_template(self, plist, ns, ctxt):
        return self.head.pp_declare_fn(self.init, self.body, plist, ctxt, ns)

class DeclFunctionHead (Ext):
    def pp_declare_fn(self, init, body, ns, ctxt, tns=None):
        self.init = init
        self.body = body
        decl = types.Declare(tns)
        for a in self.attrs:
            decl.keyword(a)
        if self.type:
            self.type.pp_dtype(decl, ns, ctxt)
        decl.astnode = (self.attrs, self.type, self.dtor)
        o = self.dtor.pp_dtor(decl, ns, ctxt)
        if not isinstance(o, types.FunctionDef):
            raise SemanticError("Function must define function object",self.src)
        if self.init:
            for i in self.init:
                io = i.id.pp_member_or_base(ns, ctxt)
                for a in i.args:
                    a.pp_expr(o.type.plist, ctxt)
        o.locals = types.Namespace(o.type.plist)
        o.astdef = self
        self.body.pp_stmt(o.locals, ctxt)
        return o

class DeclNamespaceOriginal (Ext):
    def pp_declare(self, ns, ctxt):
        o = ns.lookup(self.name)
        if isinstance(o, types.NamespaceObj):
            nns = o.nns
        else:
            nns = types.NamedNamespace(ns, self.name)
            ns.dict[self.name] = types.NamespaceObj(nns)
        for d in self.body:
            d.pp_declare(nns, ctxt)

class DeclClass (Ext):
    def pp_declare(self, ns, ctxt):
        self.pp_dtype(None, ns, ctxt)

    def pp_dtype(self, decl, ns, ctxt):
        assert self.nns is None
        if self.bases:
            bl = [b.pp_base_class(ns, ctxt) for b in self.bases]
        else:
            bl = []
        if self.name:
            name = self.name
        else:
            name = ns.anon_get_name()
        mns = types.Memberspace(ns, self.name, [b.members for b in bl])
        obj = types.Class(name, bl, mns)
        obj.astnode = self
        ns.dict[name] = obj
        for d in self.body:
            d.pp_declare(mns, ctxt)
        if decl is not None:
            decl.set_type(obj)

    def pp_template(self, plist, ns, ctxt):
        assert self.nns is None
        if self.bases:
            bl = [b.pp_base_class(plist, ctxt) for b in self.bases]
        else:
            bl = []
        if self.name:
            name = self.name
        else:
            name = ns.anon_get_name()
        mns = types.Memberspace(plist, self.name, [b.members for b in bl])
        obj = types.TemplateClass(name, bl, mns)
        ns.dict[name] = obj
        for d in self.body:
            d.pp_declare(mns, ctxt)

class DeclTemplate (Ext):
    def pp_declare(self, ns, ctxt):
        pns = types.ParamList(ns)
        for p in self.plist: p.pp_param(pns, ctxt)
        self.decl.pp_template(pns, ns, ctxt)

class DeclSimple (Ext):
    def pp_declare(self, ns, ctxt):
        decl = types.Declare()
        for a in self.attrs:
            decl.keyword(a)
        if self.type:
            self.type.pp_dtype(decl, ns, ctxt)
        for d in self.dtor:
            decl.astnode = (self.attrs, self.type, d)
            d.pp_dtor(decl, ns, ctxt)

    def pp_template(self, plist, ns, ctxt):
        if len(self.dspec) == 1 and \
                isinstance(self.dspec[0], (DeclClass, TypeClass)) \
                and not self.dtor:
            self.dspec[0].pp_template(plist, ns, ctxt)
            return
        assert False ## we don't support template fns yet!

"""
class DeclMember (Ext):
    def pp_declare(self, ns, ctxt):
        decl = types.Declare()
        
        for d in self.dspec:
            d.pp_dspecifier(decl, ns, ctxt, bool(self.dtor))
        for d in self.dtor:
            decl.astnode = (self.dspec, d)
            d.pp_dtor(decl, ns, ctxt)
"""

class DeclAccess (Ext):
    def pp_declare(self, ns, ctxt):
        pass

class ClassBase (Ext):
    def pp_base_class(self, ns, ctxt):
        if self.glbl: ns = ctxt.globalns
        if self.nns is not None: ns = self.nns.pp_nslookup(ns, ctxt)
        o = self.name.pp_lookup(ns, ctxt)
        if not isinstance(o, (types.Class, types.TemplateClass)):
            raise SemanticError("Bad base class", self.src)
        return o

    def pp_member_or_base(self, ns, ctxt):
        if self.glbl: ns = ctxt.globalns
        if self.nns is not None: ns = self.nns.pp_nslookup(ns, ctxt)
        o = self.name.pp_lookup(ns, ctxt)
        if not isinstance(o, (types.Class, types.TemplateClass)):
            raise SemanticError("Bad base class", self.src)
        return o

class IdFinaliser (Ext):
    def pp_expr(self, ns, ctxt):
        self.name.pp_lookup(ns, ctxt)

class IdMember (Ext):
    def pp_member_or_base(self, ns, ctxt):
        o = ns.lookup(self.name)
        if not isinstance(o, types.Variable):
            raise SemanticError("Bad member", self.src)
        return o

class TypeCV (Ext):
    def pp_dtype(self, decl, ns, ctxt):
        for k in self.cv:
            decl.keyword(k)
        self.t.pp_dtype(decl, ns, ctxt)

class TypeBuiltin (Ext):
    def pp_dtype(self, decl, ns, ctxt):
        for k in self.kws:
            decl.keyword(k)

class TypeAuto (Ext):
    def pp_dtype(self, decl, ns, ctxt):
        decl.keyword('auto')

class DtorInit (Ext):
    def pp_dtor(self, decl, ns, ctxt):
        if self.init: self.init.pp_expr(ns, ctxt)
        return self.d.pp_dtor(decl, ns, ctxt)

class InitList (Ext):
    def pp_expr(self, ns, ctxt):
        for v in self.data:
            v.pp_expr(ns, ctxt)

class InitConstructor (Ext):
    def pp_expr(self, ns, ctxt):
        for a in self.args:
            a.pp_expr(ns, ctxt)

class DtorArray (Ext):
    def pp_dtor(self, decl, ns, ctxt):
        return self.d.pp_dtor(decl.as_array(self.dim), ns, ctxt)

class DtorPtr (Ext):
    def pp_dtor(self, decl, ns, ctxt):
        d = self.op.pp_ptr_decl(decl, ns, ctxt)
        return self.d.pp_dtor(d, ns, ctxt)

class AbsDtorPtr (Ext):
    def pp_dtor(self, decl, ns, ctxt):
        d = self.op.pp_ptr_decl(decl, ns, ctxt)
        if self.d:
            return self.d.pp_dtor(d, ns, ctxt)

class PtrOpRef (Ext):
    def pp_ptr_decl(self, decl, ns, ctxt):
        return decl.as_ref()

class PtrOpPtr (Ext):
    def pp_ptr_decl(self, decl, ns, ctxt):
        return decl.as_ptr(self.cv)

class PtrOpQualified (Ext):
    def pp_ptr_decl(self, decl, ns, ctxt):
        if self.glbl: ns = ctxt.globalns
        ns = self.nns.pp_nslookup(ns, ctxt)
        return decl.as_ptr2m(ns, self.cv)

class DtorFunction (Ext):
    def pp_dtor(self, decl, ns, ctxt):
        pl = self.params.pp_params(ns, ctxt)
        cv = set(self.cv)
        assert self.excspec is None  # not implemented
        return self.d.pp_dtor(decl.as_func(pl, cv), ns, ctxt)

class DtorConstructor (Ext):
    def pp_dtor(self, decl, ns, ctxt):
        if self.glbl: ns = ctxt.globalns
        if self.nns is not None: ns = self.nns.pp_nslookup(ns, ctxt)
        o = self.name.pp_lookup(ns, ctxt)
        if o.name != ns.clsname:
            raise SemanticError("Wrong Constructor Name", self.src)
        return types.FunctionDef('+constructor', decl.type)

class DtorConversion (Ext):
    def pp_dtor(self, decl, ns, ctxt):
        if self.glbl: ns = ctxt.globalns
        if self.nns is not None: ns = self.nns.pp_nslookup(ns, ctxt)
        tdecl = self.name.pp_dtor(types.Declare(), ns, ctxt)
        decl.type.type = tdecl.type
        return types.FunctionDef('+conv', decl.type)

class ParamList (Ext):
    def pp_params(self, ns, ctxt):
        pns = types.ParamList(ns)
        for p in self.plist: p.pp_param(pns, ctxt)
        return pns

class Parameter (Ext):
    def pp_param(self, ns, ctxt):
        decl = types.Declare()
        self.type.pp_dtype(decl, ns, ctxt)
        if self.default is not None:
            self.default.pp_expr(ns, ctxt)
        decl.astnode = ([], self.type, self.dtor)
        if self.dtor:
            return self.dtor.pp_dtor(decl, ns, ctxt)
        return decl.declare(None, ns, ctxt, self.src)

class TypeParameter (Ext):
    def pp_param(self, ns, ctxt):
        if self.name:
            name = self.name
        else:
            name = ns.anon_get_name()
        if self.default:
            self.default.pp_type(ns, ctxt)
        obj = types.TypeParam(name)
        ns.dict[name] = obj
        return obj

#class DtorCreator (Ext):
#    def pp_dtor(self, decl, ns, ctxt):
#        if self.glbl or self.nns or not isinstance(ns, types.Memberspace) \
#                or not isinstance(decl.type, types.Function):
#            raise SemanticError("Bad Initialiser Name", self.src)
#        if self.glbl: ns = ctxt.globalns
#        if self.nns is not None: ns = self.nns.pp_nslookup(ns, ctxt)
#        o = self.name.pp_lookup(ns, ctxt)
#        if o.name != ns.clsname:
#            raise SemanticError("Wrong Initialiser Name", self.src)
#        return types.FunctionDef('+creator', decl.type)

class TypeId (Ext):
    def pp_type(self, ns, ctxt):
        decl = types.Declare()
        self.type.pp_dtype(decl, ns, ctxt)
        if self.d:
            return self.d.pp_dtor(decl, ns, ctxt)

class TypeQualified (Ext):
    def pp_dtype(self, decl, ns, ctxt):
        if self.glbl: ns = ctxt.globalns
        if self.nns is not None: ns = self.nns.pp_nslookup(ns, ctxt)
        decl.set_type(self.name.pp_lookup(ns, ctxt))

    def pp_type(self, ns, ctxt):
        if self.glbl: ns = ctxt.globalns
        if self.nns is not None: ns = self.nns.pp_nslookup(ns, ctxt)
        return self.name.pp_lookup(ns, ctxt)

class CPPInclude (ExtCPP):
    def pp_declare(self, ns, ctxt):
        if self.path:
            if self.path not in ctxt.includes:
                ctxt.includes[self.path], d = process_file(self.path, ns, ctxt)

class CPPError (ExtCPP):
    def pp_stmt(self, ns, ctxt):
        pass

    def pp_declare(self, ns, ctxt):
        pass

def process(tu, ns=None, ctxt=None):
    if ctxt is None: ctxt = Context()
    if ns is None: ns = ctxt.globalns
    tu.pp_main(ns, ctxt)
    return ctxt.globalns

def process_file(fname, ns=None, ctxt=None):
    from grammar import FileTraceSource
    from .parser import CXXTokenizer, CXXParser
    from .preproc import PreProcessor
    p = CXXParser(PreProcessor(CXXTokenizer, FileTraceSource, fname, [],
                               [(__name__, 'stdinclude')]))
    a = p.run()
    return a, process(a, ns, ctxt)

def debug_file(fname, f):
    from grammar import FileTraceSource
    from .parser import CXXTokenizer, CXXParser
    from .preproc import PreProcessor
    p = CXXParser(PreProcessor(CXXTokenizer, FileTraceSource, fname,[],
                               [(__name__, 'stdinclude')]))
    a = p.run()
    a.display(f)
