
from grammar import Buffer, EOF, Token, StringSource
from .syntax import CXXTokens
from .parser import CPPParser

import os.path
from pkg_resources import resource_exists, resource_stream

class StreamConv:

    def __init__(self, bstr):
        self.bstr = bstr

    def readline(self):
        return self.bstr.readline().decode('utf8')

class PreProcessor:

    def __init__(self, lexer, srccls, filename, include=[], resinclude=[]):
        self.lexer = lexer
        self.srccls = srccls
        if isinstance(filename, tuple):
            f = StreamConv(resource_stream(*filename))
            fn = filename[1]
        else:
            f = open(filename)
            fn = filename
        self.src = [ Buffer(lexer(srccls(f, fn))) ]
        self.fns = [ fn ]
        self.eoftkn = None
        self.locald = os.path.split(fn)[0]
        self.include = include
        self.resinclude = resinclude

    def get(self):
        t = None
        while t is None and self.src:
            t = self.do_get()
        if t is not None: return t
        return self.eoftkn

    def do_get(self):
        t = self.src[-1].get()
        if t.type == EOF:
            self.eoftkn = t
            self.src.pop()
            self.fns.pop()
            return None
        elif t.type != CXXTokens.CPP:
            return t
        return self.process(t.text.strip())

    def process(self, text):
        p = CPPParser(StringSource(text))
        r = p.run()
        if isinstance(r, CPPParser.CPPInclude):
            self.find_include(r)
            s = CXXTokens.CPP_INCLUDE
        else:
            s = CXXTokens.CPP_ERROR
        r.src = [t]
        return Token(None, s, r, src=[r])

    def find_include(self, r):
        p = os.path.normpath(os.path.join(self.locald, r.file))
        if os.path.exists(p):
            r.found_local = True
            r.resourcefile = False
            r.path = p
            return
        for pkg,d in self.resinclude:
            p = d+'/'+r.file
            if resourcse_exists(pkg, p):
                r.found_local = False
                r.path = (pkg, p)
                r.resourcefile = True
                return
        for d in self.include:
            p = os.path.normpath(os.path.join(d, r.file))
            if os.path.exists(p):
                r.found_local = False
                r.resourcefile = False
                r.path = p
                return
        r.path = None
        r.found_local = None
        r.resourcefile = None
        print("Warning -- include file not found:",r.file)
