if __name__ == '__main__':
    import os.path, sys
    cdir = sys.path[0]
    sys.path[0] = os.path.join(sys.path[0], '..')

from grammar import *

class CXXTokens (Grammar):
    DECLARE_KEYWORD = Symbol()
    TYPE_KEYWORD = Symbol()
    ACCESS_KEYWORD = Symbol()
    IDENTIFIER = Symbol()
    INTEGER_LITERAL = Symbol()
    FLOATING_LITERAL = Symbol()
    STRING_LITERAL = Symbol()
    CHARACTER_LITERAL = Symbol()
    BOOLEAN_LITERAL = Symbol()
    CPP = Symbol()
    CPP_INCLUDE = Symbol()
    CPP_ERROR = Symbol()

class CXXLexical (CXXTokens):

    token = NonTerminal()
    start = NonTerminal()

    START = (token, start)

    _ = p('start -> token')
    _ = r('start -> [ \t]*#([^\\\n]|\\\.)+', action='CPP')

    _ = r(r'token -> [ \t]+|\n', action='SPACE')
    _ = r(r'token -> //[^\n]*\n', action='SPACE')
    _ = r(r'token -> /\*([^*]|\*[^/])*\*/', action='SPACE')
#    _ = r(r'token -> \n[ \t]*#([^\\\n]|\\\.)+', action='CPP')

    _ = r(r'token -> friend|inline|mutable|register|static|explicit',
          action='DECLARE_KEYWORD')
    _ = r(r'token -> bool|char|double|float|int|long|short|signed|unsigned|'
          r'wchar_t|void', action='TYPE_KEYWORD')
    _ = r(r'token -> private|protected|public', action='ACCESS_KEYWORD')
    _ = r(r'token -> auto|const|volatile', action='LITERAL_TOKEN')
    _ = r(r'token -> false|true', action='BOOLEAN_LITERAL')

    _ = r(r'token -> break|case|catch|class|const_cast|continue|default|'
          r'delete|do|dynamic_cast|else|enum|extern|for|goto|if|'
          r'namespace|new|operator|return|sizeof|static_cast|struct|switch|'
          r'template|this|throw|try|typedef|typename|union|using|virtual|while',
          action='LITERAL_TOKEN')

    _ = r(r'token -> [A-Za-z_][A-Za-z0-9_]*', action='IDENTIFIER')
    _ = r(r'token -> ([1-9][0-9]*|0[0-7]*|0x[0-9a-fA-F]+)[uhlUHL]*',
          action='INTEGER_LITERAL')
    _ = r(r'token -> ([0-9]+\.[0-9]*|\.[0-9]+)([eE][-+]?[0-9]+)?[lfLF]?',
          action='FLOATING_LITERAL')
    _ = r(r'token -> L?"([^"\\\n]|\\[^\n])*"', action='STRING_LITERAL')
    _ = r(r"token -> L?'(\\[0-7]([0-7][0-7]?)?|\\x[0-9A-Fa-f][0-9A-Fa-f]|"
          r"[^'\\\n]|\\[^0-7x\n])'", action='CHARACTER_LITERAL')

    _ = r(r'token -> [!%&()*+,-./:;<=>?@[\]^{|}~]', action='LITERAL_TOKEN')
    _ = r(r'token -> !=|%=|&&|&=|\*=|\+\+|\+=|--|-=|->|->\*|\.\*|/=|\:\:|'
          r'<<|<=|<<=|==|>=|>>|>>=|^=|\|=|\|\|', action='LITERAL_TOKEN')

class CPPSyntax (CXXTokens):

    cpp = NonTerminal()
    path = NonTerminal()
    START = cpp

    class CPPInclude (ASTNode):
        pass

    class CPPError (ASTNode):
        pass
    
    @r(r'cpp -> #[ \t]*include[ \t]+"{path:file}"[ \t]*')
    def include_local(self, file):
        return self.CPPInclude(file=file, local=True)

    @r(r'cpp -> #[ \t]*include[ \t]+<{path:file}>[ \t]*')
    def include_std(self, file):
        return self.CPPInclude(file=file, local=False)

    @r(r'cpp -> #[ \t]*error[ \t]+[^ \t]!.!*!')
    def error_d(self, m1, msg):
        return self.CPPError(msg=m1+''.join(msg))

    @r(r'path -> [^">]!+!')
    def match_path(self, chars):
        return ''.join(chars)

class CXXSyntax (CXXTokens):

    literal = NonTerminal()
    translation_unit = NonTerminal()
    primary_expression = NonTerminal()
    id_expression = NonTerminal()
    unqualified_id = NonTerminal()
    qualified_id = NonTerminal()
    nested_name_specifier = NonTerminal()
    class_or_namespace_name = NonTerminal()
    postfix_expression = NonTerminal()
    pseudo_destructor_name = NonTerminal()
    unary_expression = NonTerminal()
    unary_operator = NonTerminal()
    new_expression = NonTerminal()
    new_placement = NonTerminal()
    new_type_id = NonTerminal()
    new_declarator = NonTerminal()
    direct_new_declarator = NonTerminal()
    new_initializer = NonTerminal()
    delete_expression = NonTerminal()
    cast_expression = NonTerminal()
    pm_expression = NonTerminal()
    multiplicative_expression = NonTerminal()
    additive_expression = NonTerminal()
    shift_expression = NonTerminal()
    relational_expression = NonTerminal()
    equality_expression = NonTerminal()
    and_expression = NonTerminal()
    exclusive_or_expression = NonTerminal()
    inclusive_or_expression = NonTerminal()
    logical_and_expression = NonTerminal()
    logical_or_expression = NonTerminal()
    conditional_expression = NonTerminal()
    assignment_expression = NonTerminal()
    assignment_operator = NonTerminal()
    expression = NonTerminal()
    constant_expression = NonTerminal()
    statement = NonTerminal()
    labeled_statement = NonTerminal()
    expression_statement = NonTerminal()
    compound_statement = NonTerminal()
    selection_statement = NonTerminal()
    condition = NonTerminal()
    iteration_statement = NonTerminal()
    for_init_statement = NonTerminal()
    jump_statement = NonTerminal()
    declaration_statement = NonTerminal()
    declaration = NonTerminal()
    block_declaration = NonTerminal()
    simple_declaration = NonTerminal()
    decl_attrib = NonTerminal()
    type_descriptor = NonTerminal()
    raw_type_descriptor = NonTerminal()
    named_type_specifier = NonTerminal()
    simple_type_specifier = NonTerminal()
    type_name = NonTerminal()
    elaborated_type_specifier = NonTerminal()
    enum_specifier = NonTerminal()
    enumerator_definition = NonTerminal()
    enumerator = NonTerminal()
    namespace_name = NonTerminal()
    namespace_definition = NonTerminal()
    named_namespace_definition = NonTerminal()
    original_namespace_definition = NonTerminal()
    unnamed_namespace_definition = NonTerminal()
    namespace_body = NonTerminal()
    namespace_alias_definition = NonTerminal()
    qualified_namespace_specifier = NonTerminal()
    using_declaration = NonTerminal()
    using_directive = NonTerminal()
    asm_definition = NonTerminal()
    linkage_specification = NonTerminal()
    init_declarator = NonTerminal()
    declarator = NonTerminal()
    direct_declarator = NonTerminal()
    ptr_operator = NonTerminal()
    cv_qualifier = NonTerminal()
    declarator_id = NonTerminal()
    cd_declarator = NonTerminal()
    cd_id = NonTerminal()
    type_id = NonTerminal()
    abstract_declarator = NonTerminal()
    direct_abstract_declarator = NonTerminal()
    parameter_declaration_clause = NonTerminal()
    parameter_declaration = NonTerminal()
    function_definition = NonTerminal()
    function_head = NonTerminal()
    function_body = NonTerminal()
    initializer = NonTerminal()
    initializer_clause = NonTerminal()
    class_name = NonTerminal()
    class_specifier = NonTerminal()
    class_head = NonTerminal()
    class_key = NonTerminal()
    member_specification = NonTerminal()
    member_declaration = NonTerminal()
    member_cd_declarator = NonTerminal()
    member_declarator = NonTerminal()
    pure_specifier = NonTerminal()
    constant_initializer = NonTerminal()
    base_clause = NonTerminal()
    base_specifier = NonTerminal()
    access_specifier = NonTerminal()
    conversion_function_id = NonTerminal()
    conversion_type_id = NonTerminal()
    conversion_declarator = NonTerminal()
    ctor_initializer = NonTerminal()
    mem_initializer = NonTerminal()
    mem_initializer_id = NonTerminal()
    operator_function_id = NonTerminal()
    operator = NonTerminal()
    template_declaration = NonTerminal()
    template_parameter = NonTerminal()
    type_parameter = NonTerminal()
    template_id = NonTerminal()
    template_name = NonTerminal()
    template_argument = NonTerminal()
    explicit_instantiation = NonTerminal()
    explicit_specialization = NonTerminal()
    try_block = NonTerminal()
    function_try_block = NonTerminal()
    handler = NonTerminal()
    exception_declaration = NonTerminal()
    throw_expression = NonTerminal()
    exception_specification = NonTerminal()

    START = translation_unit

    class DeclNode (ASTNode):
        pass

    class ExprNode (ASTNode):
        pass

    class StmtNode (ASTNode):
        pass

    class NameNode (ASTNode):
        pass

    #### A.2 Lexical conventions ####

    ## literal ##
    @p("literal -> INTEGER_LITERAL:v")
    class LiteralInt (ExprNode):
        pass

    @p("literal -> CHARACTER_LITERAL:v")
    class LiteralChar (ExprNode):
        pass

    @p("literal -> FLOATING_LITERAL:v")
    class LiteralFloat (ExprNode):
        pass

    @p("literal -> STRING_LITERAL:v")
    class LiteralStr (ExprNode):
        pass

    @p("literal -> BOOLEAN_LITERAL:v")
    class LiteralBool (ExprNode):
        pass

    #### A.3 Basic concepts ####

    ## translation-unit ##
    @p("translation_unit -> declaration!*:decls")
    class TranslationUnit (ASTNode):
        decls = ()


    #### A.4 Expressions ####

    class BinopNode (ExprNode):
        pass

    class UnopNode (ExprNode):
        pass

    ## primary-expression ##
    _ = p("primary_expression -> literal!", action='COPY')
    _ = p("primary_expression -> '(' expression! ')'", action='COPY')
    _ = p("primary_expression -> id_expression!", action='COPY')

    @p("primary_expression -> 'this'")
    class ExprThis (ExprNode):
        pass

    ## id-expression ##
    _ = p("id_expression -> unqualified_id!", action='COPY')
    _ = p("id_expression -> qualified_id!", action='COPY')

    ## unqualified-id ##
    @p("unqualified_id -> IDENTIFIER:id")
    class IdIdentifier (ASTNode):
        pass

    _ = p("unqualified_id -> operator_function_id!", action='COPY')
    _ = p("unqualified_id -> conversion_function_id!", action='COPY')
    #_ = p("unqualified_id -> template_id!", action='COPY')
    # what syntax is this rule for?  Templates are not values!

    @p("unqualified_id -> '~' class_name:name")
    class IdFinaliser (ASTNode):
        pass

    ## qualified-id ##
    @p("qualified_id -> '::'?:glbl nested_name_specifier:nns"
       " 'template'?:t unqualified_id:id")
    class IdQualified (ASTNode):
        pass

    @p("qualified_id -> '::' IDENTIFIER!")
    def qualified_id_1(self, id, src):
        return self.IdQualified(glbl=True, nns=None, t=False,
                                id=self.IdIdentifier(id=id, src=src), src=src)

    @p("qualified_id -> '::' (operator_function_id!|template_id!)")
    def qualified_id_2(self, id, src):
        return self.IdQualified(glbl=True, nns=None, t=False, id=id, src=src)

    ## nested-name-specifier ##
    @p("nested_name_specifier -> (class_name:name|namespace_name:name)"
       " '::' nested_name_specifier:nns?")
    class NestedNameSpecifier (ASTNode):
        nns = None
        t = False

    @p("nested_name_specifier -> (class_name!|namespace_name!)"
       " '::' 'template' nested_name_specifier!")
    def nested_name_specifier_1(self, name, nns, src=None):
        return self.NestedNameSpecifier(name=name, nns=nns, t=True, src=src)

    ## postfix-expression ##
    _ = p("postfix_expression -> primary_expression!", action='COPY')

    @p("postfix_expression -> postfix_expression:v '[' expression:i ']'")
    class ExprIndex (ExprNode):
        pass

    @p("postfix_expression -> postfix_expression:v '('"
       " assignment_expression!*,',':args ')'")
    class ExprCall (ExprNode):
        args = ()

    @p("postfix_expression -> simple_type_specifier:t '('"
       " assignment_expression!*,',':args ')'")
    class ExprCreate (ExprNode):
        args = ()

    @p("postfix_expression -> 'typename' '::'?:glbl nested_name_specifier:nns"
       " IDENTIFIER:name '(' assignment_expression!*,',':args ')'")
    class ExprQualifiedCreate (ExprNode):
        pass

    @p("postfix_expression -> 'typename' '::'?:glbl nested_name_specifier:nns"
       " 'template'? template_id:id '(' assignment_expression!*,',':args ')'")
    class ExprQualifiedTemplateCreate (ExprNode):
        pass

    @p("postfix_expression -> postfix_expression:v '.' 'template'?:t"
       " id_expression:id")
    class ExprAttributeAccess (ExprNode):
        pass

    @p("postfix_expression -> postfix_expression:v '->' 'template'?:t"
       " id_expression:id")
    class ExprIndirectAttributeAccess (ExprNode):
        pass

    @p("postfix_expression -> postfix_expression:v '.'"
       " pseudo_destructor_name:name")
    class ExprPseudoDestructor (ExprNode):
        pass

    @p("postfix_expression -> postfix_expression:v '->'"
       " pseudo_destructor_name:name")
    class ExprIndirectPseudoDestructor (ExprNode):
        pass

    @p("postfix_expression -> postfix_expression:v '++'")
    class ExprPostInc (UnopNode):
        pass

    @p("postfix_expression -> postfix_expression:v '--'")
    class ExprPostDec (UnopNode):
        pass

    @p("postfix_expression -> 'dynamic_cast' '<' type_id:t '>'"
       " '(' expression:v ')'")
    class ExprDynamicCast (ExprNode):
        pass

    @p("postfix_expression -> 'static_cast' '<' type_id:t '>'"
       " '(' expression:v ')'")
    class ExprStaticCast (ExprNode):
        pass

    @p("postfix_expression -> 'reinterpret_cast' '<' type_id:t '>'"
       " '(' expression:v ')'")
    class ExprReinterpretCast (ExprNode):
        pass

    @p("postfix_expression -> 'const_cast' '<' type_id:t '>'"
       " '(' expression:v ')'")
    class ExprConstCast (ExprNode):
        pass

    @p("postfix_expression -> 'typeid' '(' expression:v ')'")
    class ExprTypeId (ExprNode):
        pass

    @p("postfix_expression -> 'typeid' '(' type_id:t ')'")
    class ExprTypeIdType (ExprNode):
        pass

    ## pseudo-destructor-name ##
    @p("pseudo_destructor_name -> '::'?:glbl nested_name_specifier:nns?"
       " type_name:t '::' '~' type_name:tn")
    class PseudoDestructorNormal (ASTNode):
        pass

    @p("pseudo_destructor_name -> '::'?:glbl nested_name_specifier:nns"
       " 'template' template_id:t '::' '~' type_name:tn")
    class PseudoDestructorTemplate (ASTNode):
        pass

    @p("pseudo_destructor_name -> '::'?:glbl nested_name_specifier:nns?"
       " '~' type_name:t")
    class PseudoDestructorAnon (ASTNode):
        pass

    ## unary-expression ##
    _ = p("unary_expression -> postfix_expression!", action='COPY')

    @p("unary_expression -> '++' cast_expression:v")
    class ExprPreInc (UnopNode):
        pass

    @p("unary_expression -> '--' cast_expression:v")
    class ExprPreDec (UnopNode):
        pass

    @p("unary_expression -> unary_operator:op cast_expression:v")
    class ExprUnary (UnopNode):
        pass

    @p("unary_expression -> 'sizeof' unary_expression:v")
    class ExprSizeof (UnopNode):
        pass

    @p("unary_expression -> 'sizeof' '(' type_id:t ')'")
    class ExprSizeofType (ExprNode):
        pass

    _ = p("unary_expression -> new_expression!", action='COPY')
    _ = p("unary_expression -> delete_expression!", action='COPY')

    ## unary-operator ##
    _ = p("unary_operator -> '*'! | '&'! | '+'! | '-'! | '!'! | '~'!",
          action='COPY')

    ## new-expression ##
    @p("new_expression -> '::'?:glbl 'new' new_placement:place? new_type_id:t"
       " new_initializer:init?")
    class ExprNew (ExprNode):
        pass

    _ = p("new_expression -> '::'?:glbl 'new' new_placement:place? '(' type_id:t ')' new_initializer:init?", action='ExprNew')

    ## new-placement ##
    _ = p("new_placement -> '(' assignment_expression!+,','! ')'")

    ## new-type-id ##
    @p("new_type_id -> type_descriptor:type new_declarator:dtor?")
    class NewTypeId (ASTNode):
        pass

    ## new-declarator ##
    @p("new_declarator -> ptr_operator:op new_declarator:dtor?")
    class NewDeclaratorPtr (ASTNode):
        pass

    _ = p("new_declarator -> direct_new_declarator!", action='COPY')

    ## direct-new-declarator ##
    @p("direct_new_declarator -> direct_new_declarator:dtor?"
       " '[' expression:size ']'")
    class NewDeclaratorArray (ASTNode):
        pass

    ## new-initializer ##
    _ = p("new_initializer -> '(' assignment_expression!*,','! ')'",
          action='COPY')

    ## delete-expression ##
    @p("delete_expression -> '::'? 'delete' cast_expression:v")
    class ExprDelete (ExprNode):
        pass

    @p("delete_expression -> '::'? 'delete' '[' ']' cast_expression:v")
    class ExprDeleteArray (ExprNode):
        pass

    ## cast-expression ##
    _ = p("cast_expression -> unary_expression!", action='COPY')

    @p("cast_expression -> '(' type_id:t ')' cast_expression:v")
    class ExprCCast (ExprNode):
        pass

    ## pm-expression ##
    _ = p("pm_expression -> cast_expression!", action='COPY')

    @p("pm_expression -> pm_expression:a '.*' cast_expression:b")
    class ExprPtrToMembDereference (ExprNode):
        pass

    @p("pm_expression -> pm_expression:a '->*' cast_expression:b")
    class ExprIndirectPtrToMembDereference (ExprNode):
        pass

    ## multiplicative-expression ##
    _ = p("multiplicative_expression -> pm_expression!", action='COPY')

    @p("multiplicative_expression -> multiplicative_expression:a '*'"
       " pm_expression:b")
    class ExprMul (BinopNode):
        pass

    @p("multiplicative_expression -> multiplicative_expression:a '/'"
       " pm_expression:b")
    class ExprDiv (BinopNode):
        pass

    @p("multiplicative_expression -> multiplicative_expression:a '%'"
       " pm_expression:b")
    class ExprMod (BinopNode):
        pass

    ## additive-expression ##
    _ = p("additive_expression -> multiplicative_expression!", action='COPY')

    @p("additive_expression -> additive_expression:a '+'"
       " multiplicative_expression:b")
    class ExprAdd (BinopNode):
        pass

    @p("additive_expression -> additive_expression:a '-'"
       " multiplicative_expression:b")
    class ExprSub (BinopNode):
        pass

    ## shift-expression ##
    _ = p("shift_expression -> additive_expression!", action='COPY')

    @p("shift_expression -> shift_expression:a '<<' additive_expression:b")
    class ExprLeftShift (BinopNode):
        pass

    @p("shift_expression -> shift_expression:a '>>' additive_expression:b")
    class ExprRightShift (BinopNode):
        pass

    ## relational-expression ##
    _ = p("relational_expression -> shift_expression!", action='COPY')

    @p("relational_expression -> relational_expression:a '<'"
       " shift_expression:b")
    class ExprLT (BinopNode):
        pass

    @p("relational_expression -> relational_expression:a '>'"
       " shift_expression:b")
    class ExprGT (BinopNode):
        pass

    @p("relational_expression -> relational_expression:a '<='"
       " shift_expression:b")
    class ExprLE (BinopNode):
        pass

    @p("relational_expression -> relational_expression:a '>='"
       " shift_expression:b")
    class ExprGE (BinopNode):
        pass

    ## equality-expression ##
    _ = p("equality_expression -> relational_expression!", action='COPY')

    @p("equality_expression -> equality_expression:a '=='"
       " relational_expression:b")
    class ExprEqual (BinopNode):
        pass

    @p("equality_expression -> equality_expression:a '!='"
       " relational_expression:b")
    class ExprNotEqual (BinopNode):
        pass

    ## and-expression ##
    _ = p("and_expression -> equality_expression!", action='COPY')

    @p("and_expression -> and_expression:a '&' equality_expression:b")
    class ExprBitwiseAnd (BinopNode):
        pass

    ## exclusive-or-expression ##
    _ = p("exclusive_or_expression -> and_expression!", action='COPY')

    @p("exclusive_or_expression -> exclusive_or_expression:a '^'"
       " and_expression:b")
    class ExprBitwiseXOr (BinopNode):
        pass

    ## inclusive-or-expression ##
    _ = p("inclusive_or_expression -> exclusive_or_expression!", action='COPY')

    @p("inclusive_or_expression -> inclusive_or_expression:a '|'"
       " exclusive_or_expression:b")
    class ExprBitwiseOr (BinopNode):
        pass

    ## logical-and-expression ##
    _ = p("logical_and_expression -> inclusive_or_expression!", action='COPY')

    @p("logical_and_expression -> logical_and_expression:a '&&'"
       " inclusive_or_expression:b")
    class ExprLogicalAnd (BinopNode):
        pass

    ## logical-or-expression ##
    _ = p("logical_or_expression -> logical_and_expression!", action='COPY')

    @p("logical_or_expression -> logical_and_expression:a '||'"
       " logical_or_expression:b")
    class ExprLogicalOr (BinopNode):
        pass

    ## conditional-expression ##
    _ = p("conditional_expression -> logical_or_expression!", action='COPY')

    @p("conditional_expression -> logical_or_expression:c '?' expression:a ':'"
       " assignment_expression:b")
    class ExprCondition (ExprNode):
        pass

    ## assignment-expression ##
    _ = p("assignment_expression -> conditional_expression!", action='COPY')

    @p("assignment_expression -> logical_or_expression:lhs"
       " assignment_operator:op assignment_expression:rhs")
    class ExprAssign (ExprNode):
        pass

    _ = p("assignment_expression -> throw_expression!", action='COPY')

    ## assignment-operator ##
    _ = p("assignment_operator -> '='! | '*='! | '/='! | '%='! | '+='! |"
          " '-='! | '<<='! | '>>='! | '&='! | '^='! | '|='!", action='COPY')

    ## expression ##
    _ = p("expression -> assignment_expression!", action='COPY')

    @p("expression -> expression:a ',' assignment_expression:b")
    class ExprComma (ExprNode):
        pass

    ## constant-expression ##
    _ = p("constant_expression -> conditional_expression!", action='COPY')


    #### A.5 Statements ####

    ## statement ##
    _ = p("statement -> labeled_statement!", action='COPY')
    _ = p("statement -> expression_statement!", action='COPY')
    _ = p("statement -> compound_statement!", action='COPY')
    _ = p("statement -> selection_statement!", action='COPY')
    _ = p("statement -> iteration_statement!", action='COPY')
    _ = p("statement -> jump_statement!", action='COPY')
    _ = p("statement -> declaration_statement!", action='COPY')
    _ = p("statement -> try_block!", action='COPY')

    ## labeled-statement ##
    @p("labeled_statement -> IDENTIFIER:lbl ':' statement:st")
    class StmtLabeled (StmtNode):
        pass

    @p("labeled_statement -> 'case' constant_expression:c ':' statement:st")
    class StmtCase (StmtNode):
        pass

    @p("labeled_statement -> 'default' ':' statement:st")
    class StmtDefault (StmtNode):
        pass

    ## expression-statement ##
    @p("expression_statement -> expression:expr ';'")
    class StmtExpression (StmtNode):
        pass

    ## compound-statement ##
    @p("compound_statement -> '{' statement!*:sts '}'")
    class StmtCompound (StmtNode):
        sts = ()

    ## selection-statement ##
    @p("selection_statement -> 'if' '(' condition:cond ')' statement:thenst"
       " ( 'else' statement:elsest )?", resolve_shift='else')
    class StmtIf (StmtNode):
        elsest = None

    @p("selection_statement -> 'switch' '(' condition:cond ')' statement:st")
    class StmtSwitch (StmtNode):
        pass

    ## condition ##
    _ = p("condition -> expression!", action='COPY')

    @p("condition -> type_descriptor:type declarator:dtor '='"
       " assignment_expression:v")
    class ConditionDeclare (ASTNode):
        pass

    ## iteration-statement ##
    @p("iteration_statement -> 'while' '(' condition:cond ')' statement:st")
    class StmtWhile (StmtNode):
        pass

    @p("iteration_statement -> 'do' statement:st 'while'"
       " '(' expression:cond ')' ';'")
    class StmtDoWhile (StmtNode):
        pass

    @p("iteration_statement -> 'for' '(' for_init_statement:init"
       " condition:cond? ';' expression:inc? ')' statement:st")
    class StmtFor (StmtNode):
        cond = None
        inc = None

    ## jump-statement ##
    @p("jump_statement -> 'break' ';'")
    class StmtBreak (StmtNode):
        pass

    @p("jump_statement -> 'continue' ';'")
    class StmtContinue (StmtNode):
        pass

    @p("jump_statement -> 'return' expression:v? ';'")
    class StmtReturn (StmtNode):
        v = None

    @p("jump_statement -> 'goto' IDENTIFIER:lbl ';'")
    class StmtGoto (StmtNode):
        pass

    ## declaration-statement ##
    @p("declaration_statement -> block_declaration:decl")
    class StmtDeclaration (StmtNode):
        pass

    ## for-init-statement ##
    _ = p("for_init_statement -> expression_statement!", action='COPY')
    _ = p("for_init_statement -> simple_declaration:decl", action='StmtDeclaration')


    #### A.6 Declarations ####

    class BlockDecl (DeclNode):
        pass

    ## declaration ##
    _ = p("declaration -> block_declaration!", action='COPY')
    _ = p("declaration -> function_definition!", action='COPY')
    _ = p("declaration -> template_declaration!", action='COPY')
    _ = p("declaration -> explicit_instantiation!", action='COPY')
    _ = p("declaration -> explicit_specialization!", action='COPY')
    _ = p("declaration -> linkage_specification!", action='COPY')
    _ = p("declaration -> namespace_definition!", action='COPY')

    ## block-declaration ##
    _ = p("block_declaration -> simple_declaration!", action='COPY')
    _ = p("block_declaration -> asm_definition!", action='COPY')
    _ = p("block_declaration -> namespace_alias_definition!", action='COPY')
    _ = p("block_declaration -> using_declaration!", action='COPY')
    _ = p("block_declaration -> using_directive!", action='COPY')

    ## simple-declaration ##
    @p("simple_declaration -> decl_attrib!*:attrs type_descriptor:type "
       "init_declarator!+,',':dtor ';'")
    class DeclSimple (BlockDecl):
        type = None

    _ = p("simple_declaration -> decl_attrib!*:attrs cd_declarator!+,',':dtor"
          " ';'", action='DeclSimple')
    _ = p("simple_declaration -> (class_specifier!|enum_specifier!) ';'")
    _ = p("simple_declaration -> elaborated_type_specifier! ';'")

    ## decl-attrib
    _ = p("decl_attrib -> DECLARE_KEYWORD!")
    _ = p("decl_attrib -> 'typedef'!")
    _ = p("decl_attrib -> 'extern'!")
    _ = p("decl_attrib -> 'virtual'!")

    ## type-descriptor
    _ = p("type_descriptor -> raw_type_descriptor!")

    @p("type_descriptor -> cv_qualifier!+:cv raw_type_descriptor:t")
    class TypeCV (ASTNode):
        pass

    ## raw-type-descriptor

    @p("raw_type_descriptor -> TYPE_KEYWORD!+:kws")
    class TypeBuiltin (ASTNode):
        pass

    @p("raw_type_descriptor -> 'auto'")
    class TypeAuto (ASTNode):
        pass
    
    _ = p("raw_type_descriptor -> named_type_specifier!")
    _ = p("raw_type_descriptor -> class_specifier!")
    _ = p("raw_type_descriptor -> enum_specifier!")
    _ = p("raw_type_descriptor -> elaborated_type_specifier!")

    ## named-type-specifier
    @p("named_type_specifier -> '::'?:glbl nested_name_specifier:nns?"
       " type_name:name")
    class TypeQualified (ASTNode):
        nns = None

    @p("named_type_specifier -> '::'?:glbl nested_name_specifier:nns"
       " 'template' template_id:id")
    class TypeQualifiedTemplate (ASTNode):
        pass

    ## simple-type-sepcifier
    _ = p("simple_type_specifier -> named_type_specifier!")

    @p("simple_type_specifier -> TYPE_KEYWORD:kw")
    def simple_builtin_type(kw, src):
        return self.TypeBuiltin(kws=[kw], src=src)

    ## type-name ##
    @p("type_name -> IDENTIFIER:name")
    class NameType (NameNode):
        pass

    _ = p("type_name -> template_id!", action='COPY')

#    _ = p("type_name -> class_name!|enum_name!|typedef_name!", action='COPY')

    ## elaborated-type-specifier ##
    @p("elaborated_type_specifier -> class_key '::'?:glbl"
       " nested_name_specifier:nns? IDENTIFIER:name")
    class TypeClass (ASTNode):
        nns = None

    @p("elaborated_type_specifier -> 'enum' '::'?:glbl"
       " nested_name_specifier:nns? IDENTIFIER:name")
    class TypeEnum (ASTNode):
        pass

    @p("elaborated_type_specifier -> 'typename' '::'?:glbl"
       " nested_name_specifier:nns IDENTIFIER:name")
    class TypeTypename (ASTNode):
        pass

    @p("elaborated_type_specifier -> 'typename' '::'?:glbl"
       " nested_name_specifier:nns 'template'? template_id:id")
    class TypeTemplate (ASTNode):
        pass

    ## enum-name ##
#    @p("enum_name -> IDENTIFIER:name")
#    class NameEnum (NameNode):
#        pass

    ## enum-specifier ##
    @p("enum_specifier -> 'enum' IDENTIFIER:name?"
       " '{' enumerator_definition!*,',':elist '}'")
    class DeclEnum (ASTNode):
        pass

    ## enumerator-definition ##
    @p("enumerator_definition -> enumerator:name ('=' constant_expression:v)?")
    class Enumerator (ASTNode):
        pass

    ## enumerator ##
    _ = p("enumerator -> IDENTIFIER!", action='COPY')

    ## namespace-name ##
    @p("namespace_name -> IDENTIFIER:name")
    class NameNamespace (NameNode):
        pass
    #_ = p("namespace_name -> original_namespace_name!", action='COPY')
    #_ = p("namespace_name -> namespace_alias!", action='COPY')

    ## original-namespace-name ##
    #@p("original_namespace_name -> IDENTIFIER:name")
    #class NameNamespace (NameNode):
    #    pass

    ## namespace-definition ##
    _ = p("namespace_definition -> named_namespace_definition!", action='COPY')
    _ = p("namespace_definition -> unnamed_namespace_definition!",action='COPY')

    ## named-namespace-definition ##
    _ = p("named_namespace_definition -> original_namespace_definition!",
          action='COPY')
#    _ = p("named_namespace_definition -> extension_namespace_definition!",
#          action='COPY')

    ## original-namespace-definition ##
    @p("original_namespace_definition -> 'namespace' IDENTIFIER:name"
       " '{' namespace_body:body '}'")
    class DeclNamespaceOriginal (ASTNode):
        pass

    ## extension-namespace-definition ##
#    @p("extension_namespace_definition -> 'namespace'"
#       " original_namespace_name:ns '{' namespace_body:body '}'")
#    class DeclNamespaceExtension (ASTNode):
#        pass

    ## unnamed-namespace-definition ##
    @p("unnamed_namespace_definition -> 'namespace'"
       " '{' namespace_body:body '}'")
    class DeclNamespaceUnnamed (ASTNode):
        pass

    ## namespace-body ##
    _ = p("namespace_body -> declaration!*!", action='COPY')

    ## namespace-alias ##
    #@p("namespace_alias -> IDENTIFIER:name")
    #class NameNamespaceAlias (ASTNode):
    #    pass

    ## namespace-alias-definition ##
    @p("namespace_alias_definition -> 'namespace' IDENTIFIER:name"
       " '=' qualified_namespace_specifier:ns ';'")
    class DeclNamespaceAlias (BlockDecl):
        pass

    ## qualified-namespace-specifier ##
    @p("qualified_namespace_specifier -> '::'?:glbl nested_name_specifier:nns?"
       " namespace_name:name")
    class IdQualifiedNamespace (ASTNode):
        pass

    ## using-declaration ##
    @p("using_declaration -> 'using' 'typename'?:tn '::'?:glbl nested_name_specifier:nns unqualified_id:id ';'")
    class DeclUsingNonGlobal (BlockDecl):
        pass

    @p("using_declaration -> 'using' '::' unqualified_id:id ';'")
    class DeclUsingGlobal (BlockDecl):
        pass

    ## using-directive ##
    @p("using_directive -> 'using' 'namespace' '::'?:glbl"
       " nested_name_specifier:nns? namespace_name:name ';'")
    class DeclUsingNamespace (BlockDecl):
        pass

    ## asm-definition ##
    @p("asm_definition -> 'asm' '{' STRING_LITERAL:asm '}'")
    class DeclAssembly (BlockDecl):
        pass

    ## linkage-specification ##
    @p("linkage_specification -> 'extern' STRING_LITERAL:link"
       " '{' declaration!*:body '}'")
    class DeclLinkage (ASTNode):
        pass

    @p("linkage_specification -> 'extern' STRING_LITERAL:link declaration:decl")
    def linkage_specification_single(self, link, decl, src):
        return self.DeclLinkage(link=link, body=[decl], src=src)


    #### A.7 Declarators ####

    class DtorNode (ASTNode):
        pass

    ## init-declarator ##
    @p("init_declarator -> declarator:d initializer:init?")
    class DtorInit (DtorNode):
        init = None

    ## declarator ##
    _ = p("declarator -> direct_declarator!", action='COPY')

    @p("declarator -> ptr_operator:op declarator:d")
    class DtorPtr (DtorNode):
        pass

    ## direct-declarator ##
    _ = p("direct_declarator -> declarator_id!", action='COPY')

    @p("direct_declarator -> direct_declarator:d"
       " '(' parameter_declaration_clause:params ')'"
       " cv_qualifier!*:cv exception_specification:excspec?")
    class DtorFunction (DtorNode):
        cv = ()

    @p("direct_declarator -> direct_declarator:d"
       " '[' constant_expression:dim? ']'")
    class DtorArray (ASTNode):
        dim = None

    _ = p("direct_declarator -> '(' declarator! ')'")

    ## ptr-operator ##
    @p("ptr_operator -> '*' cv_qualifier!*:cv")
    class PtrOpPtr (ASTNode):
        pass

    @p("ptr_operator -> '&'")
    class PtrOpRef (ASTNode):
        pass

    @p("ptr_operator -> '::'?:glbl nested_name_specifier:nns '*'"
       " cv_qualifier!*:cv")
    class PtrOpQualified (ASTNode):
        pass

    ## cv-qualifier ##
    
    _ = p("cv_qualifier -> 'const'!")
    _ = p("cv_qualifier -> 'volatile'!")

    ## declarator-id ##
    _ = p("declarator_id -> id_expression!")

    ## cd-declarator ##
    _ = p("cd_declarator -> cd_id:d '(' parameter_declaration_clause:params ')'"
          " exception_specification:excspec?", action='DtorFunction')
    
    ## cd-id ##
    @p("cd_id -> '::'?:glbl nested_name_specifier:nns? type_name:name")
    class DtorConstructor (DtorNode):
        pass

    @p("cd_id -> '::'?:glbl nested_name_specifier:nns? '~' type_name:name")
    class DtorDestructor (DtorNode):
        pass

    @p("cd_id -> '::'?:glbl nested_name_specifier:nns?"
       " conversion_function_id:name")
    class DtorConversion (DtorNode):
        pass

    ## type-id ##
    @p("type_id -> type_descriptor:type abstract_declarator:d?")
    class TypeId (ASTNode):
        pass

    ## abstract-declarator ##

    class AbsDtorNode (ASTNode):
        pass

    @p("abstract_declarator -> ptr_operator:op abstract_declarator:d?")
    class AbsDtorPtr (AbsDtorNode):
        pass

    _ = p("abstract_declarator -> direct_abstract_declarator!")

    ## direct-abstract-declarator ##
    @p("direct_abstract_declarator -> direct_abstract_declarator:d?"
       " '(' parameter_declaration_clause:params ')' cv_qualifier!*:cv"
       " exception_specification:excspec?")
    class AbsDtorFunction (ASTNode):
        pass

    @p("direct_abstract_declarator -> direct_abstract_declarator:d?"
       " '[' constant_expression:dim? ']'")
    class AbsDtorArray (ASTNode):
        pass

    _ = p("direct_abstract_declarator -> '(' abstract_declarator! ')'")

    ## parameter-declaration-clause ##
    @p("parameter_declaration_clause -> parameter_declaration!*,',':plist"
       " '...'?:etc")
    class ParamList (ASTNode):
        plist = ()

    @p("parameter_declaration_clause -> parameter_declaration!+,',':plist"
       " ',' '...'")
    def parameter_declaration_clause_1(self, plist, src):
        return self.Params(plist=plist, etc=True, src=src)

    ## parameter-declaration ##
    @p("parameter_declaration -> type_descriptor:type declarator:dtor")
    class Parameter (ASTNode):
        default=None
        dtor=None

    _ = p("parameter_declaration -> type_descriptor:type declarator:dtor"
          " '=' assignment_expression:default", action='Parameter')

    _ = p("parameter_declaration -> type_descriptor:type"
          " abstract_declarator:dtor?", action='Parameter')

    _ = p("parameter_declaration -> type_descriptor:type"
          " abstract_declarator:dtor? '=' assignment_expression:default",
          action='Parameter')

    ## function-definition ##
    @p("function_definition -> function_head:head ctor_initializer:init?"
       " function_body:body")
    class DeclFunction (ASTNode):
        init = ()
    
    _ = p("function_definition -> function_head:head function_try_block:tryb",
          action="DeclFunction")

    ## function-head ##
    @p("function_head -> decl_attrib!*:attrs cd_declarator:dtor")
    class DeclFunctionHead (ASTNode):
        type = None

    _ = p("function_head -> decl_attrib!*:attrs type_descriptor:type"
          " declarator:dtor", action='DeclFunctionHead')

   ## function-body ##
    _ = p("function_body -> compound_statement!", action='COPY')

    ## initializer ##
    _ = p("initializer -> '=' initializer_clause!", action='COPY')

    @p("initializer -> '(' assignment_expression!+,',':args ')'")
    class InitConstructor (ASTNode):
        pass

    ## initializer-clause ##
    _ = p("initializer_clause -> assignment_expression!", action='COPY')

    @p("initializer_clause -> '{' initializer_clause!+,',':data ','? '}'")
    class InitList (ASTNode):
        pass

    @p("initializer_clause -> '{' '}'")
    def initializer_clause_2(self, src):
        return InitList(data=[], src=src)


    #### A.8 Classes ####

    ## class-name ##
    @p("class_name -> IDENTIFIER:name")
    class NameClass (NameNode):
        pass

    _ = p("class_name -> template_id!", action='COPY')

    ## class-specifier ##
    @p("class_specifier -> class_head!! '{' member_specification!*:body '}'")
    class DeclClass (ASTNode):
        name = None
        nns = None

    ## class-head ##
    _ = p("class_head -> class_key:t base_clause:bases?", action='DICT')

    _ = p("class_head -> class_key:t nested_name_specifier:nns? IDENTIFIER:name"
          " base_clause:bases?", action='DICT')

    _ = p("class_head -> class_key:t nested_name_specifier:nns? template_id:tid"
       " base_clause:bases?", action='DICT')

    ## class-key ##
    _ = p("class_key -> 'class' | 'struct' | 'union'")

    ## member-specification ##
    _ = p("member_specification -> member_declaration")

    @p("member_specification -> access_specifier:access ':'")
    class DeclAccess (ASTNode):
        pass

    ## member-declaration ##
    _ = p("member_declaration -> decl_attrib!*:attrs type_descriptor:type "
          " member_declarator!*,',':dtor ';'", action='DeclSimple')

    _ = p("member_declaration -> decl_attrib!*:attrs"
          " member_cd_declarator!+,',':dtor ';'", action='DeclSimple')

 #   _ = p("member_declaration -> (class_specifier!|enum_specifier!) ';'")
    _ = p("member_declaration -> elaborated_type_specifier! ';'")
    _ = p("member_declaration -> function_definition! ';'?")

    @p("member_declaration -> '::'?:glbl nested_name_specifier:nns"
       " 'template'?:t unqualified_id:id ';'")
    class MemberId (ASTNode):
        pass

    _ = p("member_declaration -> using_declaration")
    _ = p("member_declaration -> template_declaration")

    ## member-cd-declarator ##
    @p("member_cd_declarator -> cd_declarator:d pure_specifier")
    class DtorPure (DtorNode):
        pass

    _ = p("member_cd_declarator -> cd_declarator")
    
    ## member-declarator ##
    @p("member_declarator -> declarator:d pure_specifier")
    class DtorPure (DtorNode):
        pass

    _ = p("member_declarator -> declarator:d constant_initializer:init?",
          action='DtorInit')

    @p("member_declarator -> IDENTIFIER:name? ':' constant_expression:size")
    class DtorSize (DtorNode):
        pass

    ## pure-specifier ##
    @p("pure_specifier -> '=' INTEGER_LITERAL:i")
    def pure_specifier_0(self, i, src):
        if i != 0:
            raise ParseError("Pure specifier must be 0", src)
        return None

    ## constant-initializer ##
    _ = p("constant_initializer -> '=' constant_expression!", action='COPY')

    #### A.9 Derieved classes ####

    ## base-clause ##
    _ = p("base_clause -> ':' base_specifier!+,','!", action='COPY')

    ## base-specifier ##
    @p("base_specifier -> '::'?:glbl nested_name_specifier:nns?"
       " class_name:name")
    class ClassBase (ASTNode):
        pass

    @p("base_specifier -> 'virtual' access_specifier:acc? '::'?:glbl"
       " nested_name_specifier:nns? class_name:name")
    def base_specifier_1(self, **kwds):
        return self.ClassBase(v=True, **kwds)

    _ = p("base_specifier -> access_specifier:acc 'virtual'?:v '::'?:glbl"
          " nested_name_specifier:nns? class_name:name", action='ClassBase')

    ## access-specifier ##

    _ = p("access_specifier -> ACCESS_KEYWORD!", action='COPY')

    #### A.10 Special member functions ####

    ## conversion-function-id ##
    _ = p("conversion_function_id -> 'operator' conversion_type_id!",
          action='COPY')

    ## conversion-type-id ##
    @p("conversion_type_id -> type_descriptor:type"
       " conversion_declarator:dtor?")
    class IdConversionFn (ASTNode):
        pass

    ## conversion-declarator ##
    _ = p("conversion_declarator -> ptr_operator:op conversion_declarator:d?",
          action='AbsDtorPtr')

    ## ctor-initializer ##
    _ = p("ctor_initializer -> ':' mem_initializer!+,','!", action='COPY')

    ## mem-initializer ##
    @p("mem_initializer -> mem_initializer_id:id"
       " '(' assignment_expression!*,',':args ')'")
    class MemInitializer (ASTNode):
        args = ()

    ## mem-initializer-id ##
    _ = p("mem_initializer_id -> '::'?:glbl nested_name_specifier:nns?"
          " class_name:name", action='ClassBase')

    @p("mem_initializer_id -> IDENTIFIER:name")
    class IdMember (ASTNode):
        pass


    #### A.11 Overloading ####

    ## operator-function-id ##
    @p("operator_function_id -> 'operator' operator:op")
    class IdOperatorFn (ASTNode):
        pass

    ## operator ##
    _ = p("operator -> 'new'! | 'delete'! | '+'! | '-'! | '*'! | '/'! | '%'! |"
          " '^'! | '&'! | '|'! | '~'! | '!'! | '='! | '<'! | '>'! | '+='! |"
          " '-='! | '*='! | '/='! | '%='! | '^='! | '&='! | '|='! | '<<'! |"
          " '>>'! | '<<='! | '>>='! | '=='! | '!='! | '<='! | '>='! | '&&'! |"
          " '||'! | '++'! | '--'! | ','! | '->*'! | '->'!", action='COPY')

    @p("operator -> 'new' '[' ']'")
    def operator_new_arr(self, src):
        return 'new[]'

    @p("operator -> 'delete' '[' ']'")
    def operator_delete_arr(self, src):
        return 'delete[]'

    @p("operator -> '(' ')'")
    def operator_call(self, src):
        return '()'

    @p("operator -> '[' ']'")
    def operator_index(self, src):
        return '[]'


    #### A.12 Templates ####

    ## template-declaration ##
    @p("template_declaration -> 'export'?:exp 'template' '<' template_parameter!+,',':plist '>' declaration:decl")
    class DeclTemplate (ASTNode):
        pass

    ## template-parameter ##
    _ = p("template_parameter -> type_parameter!", action='COPY')
    _ = p("template_parameter -> parameter_declaration!", action='COPY')

    ## type-parameter ##
    @p("type_parameter -> ('class'|'typename') IDENTIFIER:name?"
       " ( '=' type_id:default )?")
    class TypeParameter (ASTNode):
        name = None
        default = None

    @p("type_parameter -> 'template' '<' template_parameter!+,',':plist '>'"
       " 'class' IDENTIFIER:name? ( '=' id_expression:default )?")
    class TypeParamTemplate (ASTNode):
        pass

    ## template-id ##
    @p("template_id -> template_name:name '<' template_argument!*,',':args '>'")
    class IdTemplate (NameNode):
        pass

    ## template-name ##
    _ = p("template_name -> IDENTIFIER!", action='COPY')

    ## template-argument ##
    @p("template_argument -> assignment_expression:v")
    class TmpltArgValue (ASTNode):
        pass

    @p("template_argument -> type_id:t")
    class TmpltArgType (ASTNode):
        pass

    #_ = p("template_argument -> id_expression!", action='COPY')

    ## explicit-instantiation ##
    @p("explicit_instantiation -> 'template' declaration:decl")
    class DeclExplicitInst (ASTNode):
        pass

    ## explicit-specialization ##
    @p("explicit_specialization -> 'template' '<' '>' declaration:decl")
    class DeclExplicitSpec (ASTNode):
        pass


    #### A.13 Exception handling ####

    ## try-block ##
    @p("try_block -> 'try' compound_statement:body handler!+:handlers")
    class StmtTry (ASTNode):
        pass

    ## function-try-block ##
    @p("function_try_block -> 'try' ctor_initializer:init?"
       " function_body:body handler!+:handlers")
    class FuncTryBlock (ASTNode):
        pass

    ## handler ##
    @p("handler -> 'catch' '(' exception_declaration:exc ')'"
       " compound_statement:body")
    class StmtCatch (ASTNode):
        pass

    ## exception-declaration ##
    @p("exception_declaration -> type_descriptor:type ( declarator:dtor |"
       " abstract_declarator:dtor )?")
    class DeclException (ASTNode):
        dtor = None

    @p("exception_declaration -> '...'")
    class DeclExceptionAny (ASTNode):
        pass

    ## throw-expression ##
    @p("throw_expression -> 'throw' assignment_expression:v?")
    class ExprThrow (ExprNode):
        pass

    ## exception-specification ##
    @p("exception_specification -> 'throw' '(' type_id!*',':xlist ')'")
    class ExceptionSpecification (ASTNode):
        pass

    ## #include
    _ = p("declaration -> CPP_INCLUDE!", action='COPY')

    ## #error
    _ = p("statement -> CPP_ERROR!", action='COPY')
    _ = p("declaration -> CPP_ERROR!", action='COPY')

if __name__ == '__main__':
    CXXTokenizer = RegularParserGenerator(CXXLexical, "CXXTokenizer",
                                          ConflictFirstListed,
                                          SourceTrace, LexicalAnalyser).parser()
    CPPParser = LALRParserGenerator(CPPSyntax, "CPPParser").parser()
    CXXParser = GLRParserGenerator(CXXSyntax, "CXXParser", Reproduce).parser()
    w = PythonWriter(os.path.join(cdir, "parser.py"), '.syntax')
    w.add(CXXTokenizer)
    w.add(CPPParser)
    w.add(CXXParser)
    w.write("", """
for n,s in enumerate(CXXParser.TABLE.states): s.n = n
""")
